
import numpy as np

import Sofa
import Sofa.Core
import Sofa.Simulation
import SofaRuntime
from splib3.animation.animate import Animation
from datetime import datetime


SofaRuntime.importPlugin("Sofa.Component")


class RewardShaper(Sofa.Core.Controller):
	"""Compute the reward.

	Methods:
	-------
		__init__: Initialization of all arguments.
		getReward: Compute the reward.
		update: Initialize the value of cost.

	Arguments:
	---------
		rootNode: <Sofa.Core>
			The scene.
		goal_pos: coordinates
			The position of the goal.
		effMO: <MechanicalObject>
			The mechanical object of the element to move.
		cost:
			Evolution of the distance between object and goal.

	"""
	def __init__(self, *args, **kwargs):
		"""Initialization of all arguments.

		Parameters:
		----------
			kwargs: Dictionary
				Initialization of the arguments.

		Returns:
		-------
			None.

		"""
		Sofa.Core.Controller.__init__(self, *args, **kwargs)

		self.rootNode = None
		if kwargs["rootNode"]:
			self.root = kwargs["rootNode"]
		self.goal_pos = None
		if kwargs["goalPos"]:
			self.goal_pos = kwargs["goalPos"]

		self.init_dist = None
		self.prev_dist = None

	def getReward(self):
		"""Compute the reward.

		Parameters:
		----------
			None.

		Returns:
		-------
			The reward and the cost.

		"""
		tip = self.root.InstrumentCombined.DOFs.position[-1][:3]
		current_dist = -np.sqrt(np.sum(np.square(np.array(tip)-np.array(self.goal_pos))))
		
		#reward = (self.prev_dist - current_dist)/self.prev_dist
		#self.prev_dist = current_dist
		#time = self.root.time.value

		interim = [-16, 18,126]

		interim_dist = -np.sqrt(np.sum(np.square(np.array(tip)-np.array(interim))))
		
		if current_dist > -12:
			reward = 1000
			with open(f'./sofagym/envs/Cath_2img/goal_hit.txt', 'a') as f:
				f.write('goal hit, ' + str(datetime.now()) +'\n')
		elif interim_dist > -4 or current_dist>-70:
			reward = current_dist + 10
		else:
			reward = current_dist+interim_dist-100
			

		return reward, reward

	def update(self):
		"""Update function.

		This function is used as an initialization function.

		Parameters:
		----------
			None.

		Arguments:
		---------
			None.

		"""

		tip = self.root.InstrumentCombined.DOFs.position[-1][:3]
		self.init_dist = np.linalg.norm(np.array(tip)-np.array(self.goal_pos))
		self.prev_dist = self.init_dist


class GoalSetter(Sofa.Core.Controller):
	"""Compute the goal.

	Methods:
	-------
		__init__: Initialization of all arguments.
		update: Initialize the value of cost.

	Arguments:
	---------
		goalMO: <MechanicalObject>
			The mechanical object of the goal.
		goalPos: coordinates
			The coordinates of the goal.

	"""

	def __init__(self, *args, **kwargs):
		"""Initialization of all arguments.

		Parameters:
		----------
			kwargs: Dictionary
				Initialization of the arguments.

		Returns:
		-------
			None.

		"""
		Sofa.Core.Controller.__init__(self, *args, **kwargs)

		self.rootNode = None
		if kwargs["rootNode"]:
			self.rootNode = kwargs["rootNode"]
		self.goalMO = None
		if kwargs["goalMO"]:
			self.goalMO = kwargs["goalMO"]
		self.goalPos = None
		if kwargs["goalPos"]:
			self.goalPos = kwargs["goalPos"]

	def update(self):
		"""Set the position of the goal.

		This function is used as an initialization function.

		Parameters:
		----------
			None.

		Arguments:
		---------
			None.

		"""
		with self.goalMO.position.writeable() as position:
			print("update", self.goalPos)
			position[0] = self.goalPos

	def set_mo_pos(self, goal):
		"""Modify the goal.

		Not used here.
		"""
		pass


def _getGoalPos(root):
	"""Get XYZ position of the goal.

	Parameters:
	----------
		rootNode: <Sofa.Core>
			The scene.

	Returns:
	-------
		The position of the goal.
	"""
	return root.Goal.GoalMO.position[0]


def getState(root):
	"""Compute the state of the environment/agent.

	Parameters:
	----------
		rootNode: <Sofa.Core>
			The scene.

	Returns:
	-------
		State: list of float
			The state of the environment/agent.
	"""
	xtips = []
	rotations = []

	for instrument in range(1):
		xtips.append(root.InstrumentCombined.m_ircontroller.xtip.value[instrument].tolist())
		rotations.append(root.InstrumentCombined.m_ircontroller.rotationInstrument.value[instrument].tolist())

	tip = root.InstrumentCombined.DOFs.position[-1][:3].tolist()
	
	node_3 = root.InstrumentCombined.DOFs.position[-3][:3].tolist()
	node_7 = root.InstrumentCombined.DOFs.position[-7][:3].tolist()
	node_11 = root.InstrumentCombined.DOFs.position[-11][:3].tolist()
	node_15 = root.InstrumentCombined.DOFs.position[-15][:3].tolist()
	node_19 = root.InstrumentCombined.DOFs.position[-19][:3].tolist()

	goal_pos = _getGoalPos(root).tolist()

	state = tip
	
	#print('state:',state)

	return state


def getReward(root):
	"""Compute the reward using Reward.getReward().

	Parameters:
	----------
		rootNode: <Sofa.Core>
			The scene.

	Returns:
	-------
		done, reward

	"""
	
	reward, cost = root.Reward.getReward()

	if reward == 1000:
		print('Reached Goal!!!!!!!!!!!!!!!!!!!!')
		return True, 1000

	return False, reward


def get_ircontroller_state(node, instrument=0):
	"""
	Get state (translation, rotation) of th Interventional Radiology Controller
	"""
	return [float(node.m_ircontroller.xtip.value[instrument]),
			float(node.m_ircontroller.rotationInstrument.value[instrument])]


def startCmd(root, action, duration):
	"""Initialize the command from root and action.

	Parameters:
	----------
		rootNode: <Sofa.Core>
			The scene.
		action: int
			The action.
		duration: float
			Duration of the animation.

	Returns:
	------
		None.

	"""
	scale = int(duration/0.01 + 1)
	controlled_instrument, cmd_translation, cmd_rotation = action_to_command(action, scale)
	source = get_ircontroller_state(root.InstrumentCombined, instrument=controlled_instrument)
	target_translation = source[0] + cmd_translation
	target = [target_translation if target_translation > 0 else 0.1, source[1] + cmd_rotation]
	start_cmd(root, root.InstrumentCombined, source, target, duration, controlled_instrument)


def displace(root, displacement, n_steps):
	"""Change the value of the angle.

	Parameters:
	----------
		acuator:
			The motor we consider.
		displacement: int
			The increment for the angle.

	Returns:
	-------
		None.
	"""

	cable_num, cable_disp = displacement
	cables = [root.Robot.Actuators.north, root.Robot.Actuators.west, root.Robot.Actuators.south,
			  root.Robot.Actuators.east]
	cable = cables[cable_num - 1]
	cable.value += cable_disp/n_steps


def start_cmd(rootNode, IRC_node, source, target, duration, instrument=0):
	def execute_animation(controller, anim_source, anim_target, factor, anim_instrument):
		"""
		Execute animation on the IRC to go from source to target
		"""
		with controller.xtip.writeable() as xtip:
			xtip[anim_instrument] = anim_source[0] + (anim_target[0] - anim_source[0]) * factor
		if anim_instrument == 0:
			with controller.rotationInstrument.writeable() as rotation:
				rotation[0] = anim_source[1] + (anim_target[1] - anim_source[1]) * factor

	rootNode.AnimationManager.addAnimation(
		Animation(
			onUpdate=execute_animation,
			params={"controller": IRC_node.m_ircontroller,
					"anim_source": source,
					"anim_target": target,
					"anim_instrument": instrument},
			duration=duration, mode="once"))

	return


def action_to_command(action, scale):
	"""Link between Gym action (int) and SOFA command (displacement of cables).

	Parameters:
	----------
		action: int
			The number of the action (Gym).

	Returns:
	-------
		The command (number of the cabl and its displacement).
	"""
	if action == 0:
		controlled_instrument = 0
		cmd_translation = 2.0 * scale / 2.0
		cmd_rotation = 0.0
	elif action == 1:
		controlled_instrument = 0
		cmd_translation = 0.0
		cmd_rotation = 1/15 * scale / 2.0
	elif action == 2:
		controlled_instrument = 0
		cmd_translation = 0.0
		cmd_rotation = -1/15 * scale / 2.0
	elif action == 3:
		controlled_instrument = 0
		cmd_translation = -0.7 * scale / 2.0
		cmd_rotation = 0.0

	else:
		raise NotImplementedError("Action is not in range 0 - 3")

	return controlled_instrument, cmd_translation, cmd_rotation


def getPos(root):
	"""Retun the position of the mechanical object of interest.

	Parameters:
	----------
		root: <Sofa root>
			The root of the scene.

	Returns:
	-------
		_: list
			The position(s) of the object(s) of the scene.
	"""
	return


def setPos(root, pos):
	"""Set the position of the mechanical object of interest.

	Parameters:
	----------
		root: <Sofa root>
			The root of the scene.
		pos: list
			The position(s) of the object(s) of the scene.

	Returns:
	-------
		None.

	Note:
	----
		Don't forget to init the new value of the position.

	"""
	return
